from flask import Flask, jsonify
from flask_restful import Api
from sqlalchemy.exc import IntegrityError
from simplyrestful.exceptions import NotFound, Conflict


app = Flask(__name__)

handle_exceptions = app.handle_exception
handle_user_exception = app.handle_user_exception

api = Api(app)

app.handle_user_exception = handle_exceptions
app.handle_user_exception = handle_user_exception

app.url_map.strict_slashes = False


@app.errorhandler(NotFound)
def not_found_error(e):
    return jsonify(dict(message='Not found', detail=str(e))), 404


@app.errorhandler(Conflict)
def not_found_error(e):
    return jsonify(dict(message='Conflict', detail=str(e))), 409


@app.errorhandler(IntegrityError)
def integrity_error(e):
    return jsonify(dict(message='Integrity error', detail=str(e))), 409


@app.errorhandler(Exception)
def unknown_error(e):
    return jsonify(dict(message='Unknown error', detail=str(e))), 500


@app.teardown_appcontext
def shutdown_session(exception=None):
    from database import session
    session.close()



from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.ext.declarative import declarative_base

from settings import settings


engine = create_engine(settings['DATABASE'])

session = scoped_session(sessionmaker(bind=engine))

Base = declarative_base()
Base.query = session.query_property()
Base.metadata.bind = engine



from importlib import import_module


def instantiate(class_string):
    module_name, class_name = class_string.rsplit('.', 1)
    try:
        module_ = import_module(module_name)
        try:
            class_ = getattr(module_, class_name)()
        except AttributeError:
            raise Exception('Class does not exist')
    except ImportError:
        raise Exception('Module does not exist')
    return class_ or None



from datetime import datetime

from simplyrestful.database import session
from simplyrestful.filtering import Filter
from simplyrestful.models import get_class_by_table_name
from simplyrestful.exceptions import NotFound
from simplyrestful.util import instantiate

from settings import settings


class Serializer(object):
    authenticators = []
    authorizers = []
    validators = []
    fields = None

    methods = {
        datetime: lambda x: x.strftime(settings['DATE_FORMAT'])
    }

    @property
    def model(self):
        raise NotImplementedError()

    def __init__(self):
        self.query = session.query(self.model)
        if not self.authenticators:
            self.authenticators = [
                instantiate(a) for a in settings['DEFAULT_AUTHENTICATION']
            ]
        if not self.authorizers:
            self.authorizers = [
                instantiate(a) for a in settings['DEFAULT_AUTHORIZATION']
            ]
        self.user = self.authenticate()

    def create(self, data):
        try:
            instance = self.model()
            self.validate(data)
            self.deserialize(data, instance)
            session.add(instance)
            session.flush()
            serialized = self.serialize(instance)
            session.commit()
            return serialized
        except:
            session.rollback()
            raise

    def update(self, identifier, data):
        try:
            instance = self._get_instance(identifier)
            self.validate(data, instance=instance)
            self.deserialize(data, instance)
            session.flush()
            serialized = self.serialize(instance)
            session.commit()
            return serialized
        except:
            session.rollback()
            raise

    def read(self, identifier):
        try:
            return self.serialize(self._get_instance(identifier))
        except:
            session.rollback()
            raise

    def _get_instance(self, identifier):
        instance = self.query.get(identifier)
        if not instance:
            raise NotFound(
                'Identifier "{}"'.format(identifier)
            )
        return instance

    def list(self, filtering):
        try:
            filters = Filter(self.model, filtering)

            query = self.query.join(
                * filters.joins
            ).filter(
                * self._list_filters()
            ).filter(
                filters.orm_filters
            ).order_by(
                * filters.order_by
            )

            return dict(
                results=[
                    self.serialize(m)
                    for m in query.limit(filters.limit).offset(filters.offset).all()
                ],
                count=query.count()
            )
        except:
            session.rollback()
            raise

    def _list_filters(self):
        return ()

    def delete(self, identifier):
        try:
            self._delete(identifier)
            session.commit()
        except:
            session.rollback()
            raise

    def _delete(self, identifier):
        self.query.filter_by(id=identifier).delete()

    def serialize(self, instance):
        serialized = dict()
        for prop in instance.__table__.columns:
            name = prop.name
            if not self.fields or name in self.fields:
                value = getattr(instance, name)
                value_type = type(value)

                if value_type in self.methods:
                    value = self.methods[value_type](value)

                relationships = instance.relationships
                if name in relationships and getattr(self, relationships[name], None):
                    # Consider that it may be not set
                    value = getattr(self, relationships[name])().serialize(
                        getattr(instance, relationships[name])
                    )
                    name = relationships[name]

                serialized[name] = value
        return serialized

    def deserialize(self, data, instance):
        data.update(
           Serializer.deserialize_relationships(data, instance)
        )
        for prop in data:
            if prop not in instance.primary_keys:
                setattr(instance, prop, data.get(prop))

    @staticmethod
    def deserialize_relationships(data, instance):
        for relationship in instance.__mapper__.relationships:
            target = relationship.local_remote_pairs[0][1]
            key = relationship.key
            if data.get(key):
                fields = data.get(key)
                target_class = get_class_by_table_name(target.table.name)

                # TODO: Remove
                fields = {k: v for k, v in fields.iteritems() if type(v) is not dict}

                value = session.query(target_class).filter_by(**fields).one()
                data[key] = value
        return data

    def authenticate(self):
        for authenticator in self.authenticators:
            user = authenticator.authenticate()
            if user:
                return user
        raise Exception('Authentication error')

    def authorize(self, instance):
        for authorizer in self.authorizers:
            authorizer.authorize(instance)

    def validate(self, data, instance=None):
        for validator in self.validators:
            validator().validate(data, instance=instance)



from sqlalchemy import and_

from settings import settings


class Filter(object):

    map = {
        'eq': '__eq__',
        'gt': '__gt__',
        'ge': '__ge__',
        'lt': '__lt__',
        'le': '__le__',
        'in': 'in_',
        'notin': 'notin_',
        'like': 'like',
        'notlike': 'notlike',
        'ilike': 'ilike',
        'notilike': 'notilike',
        'is': 'is_',
        'isnot': 'isnot',
        'intersects': 'ST_Intersects',
        'contains': 'ST_Intersects'
    }

    multiple = ['in', 'notin']  # TODO: Implement Between filter

    def __init__(self, model, filters):
        self.model = model
        self.filters = filters

        self._orm_filters = []
        self.joins = []
        self.order_by = []

        size = int(self.filters.pop('limit', settings['DEFAULT_PAGE_SIZE']))

        self.limit = settings['MAX_PAGE_SIZE'] if size > settings['MAX_PAGE_SIZE'] else size
        self.offset = int(self.filters.pop('offset', 0))

        self._translate_order()
        self._translate_filters()

    @property
    def orm_filters(self):
        return and_(*self._orm_filters)

    def _translate_filters(self):
        for f, value in self.filters.iteritems():
            split = f.split('__')

            last = split[-1]

            op = split.pop() if last in self.map else 'eq'

            column_name = split.pop()

            model = self.model
            for nested in split:
                model = getattr(model, nested).mapper.class_
                if model not in self.joins:
                    self.joins.append(model)

            if op in self.multiple:
                value = value.split(';')
                if not value[-1]:
                    value.pop()

            self._orm_filters.append(
                getattr(
                    getattr(model, column_name),
                    self.map[op]
                )(value)
            )

    def _translate_order(self):
        ops = ['asc', 'desc']
        order = self.filters.pop('order_by', None)
        orders = order.split(';') if order else []
        for order in orders:
            split = order.split('__')
            last = split[-1]

            op = split.pop() if last in ops else 'asc'

            column_name = split.pop()

            model = self.model
            for nested in split:
                model = getattr(model, nested).mapper.class_
                self.joins.append(model)
            self.order_by.append(
                getattr(getattr(model, column_name), op)()
            )



from sqlalchemy.sql import ClauseElement
from geometry import Geometry
from model import Model
from simplyrestful.database import Base


def get_class_by_table_name(table_name):
  for c in Base._decl_class_registry.values():
    if hasattr(c, '__tablename__') and c.__tablename__ == table_name:
      return c


def get_or_create(session, model, defaults=None, **kwargs):
    instance = session.query(model).filter_by(**kwargs).first()
    if instance:
        return instance, False
    else:
        params = dict((k, v) for k, v in kwargs.iteritems() if not isinstance(v, ClauseElement))
        params.update(defaults or {})
        instance = model(**params)
        session.add(instance)
        return instance, True





settings = dict(
    DATABASE='postgresql://postgres:postgres@localhost/simply_restful',
    HOST='0.0.0.0',
    PORT=5000,
    DATE_FORMAT='%Y-%m-%dT%H:%M:%S.%fZ',
    DEFAULT_AUTHENTICATION=['simplyrestful.authenticators.NullAuthenticator'],
    DEFAULT_AUTHORIZATION=[],
    DEFAULT_PAGE_SIZE=10,
    MAX_PAGE_SIZE=100
)


def configure_from_module(user_settings_module_name):
    import importlib

    mod = importlib.import_module(user_settings_module_name)

    user_defined_settings = dict(
        [
            (name, cls)
            for name, cls
            in mod.__dict__.items()
            if not name.startswith('_')
        ]
    )

    settings.update(user_defined_settings)



import json
from geoalchemy2 import Geometry as GeoAlchemyGeometry
import geoalchemy2.shape
from shapely.geometry import shape, mapping
import geojson


class Geometry(GeoAlchemyGeometry):
    def result_processor(self, dialect, coltype):
        super_proc = super(Geometry, self).result_processor(dialect, coltype)

        def process(v):
            return v and mapping(
                geoalchemy2.shape.to_shape(super_proc(v))
            )

        return process

    def bind_processor(self, dialect):
        super_proc = super(Geometry, self).bind_processor(dialect)

        def process(v):
            return super_proc(
                'SRID=%d;%s' % (self.srid, shape(geojson.loads(json.dumps(v))).wkt)
                if isinstance(v, dict) else v
            )

        return process



from flask import request
from flask_restful import Resource as FlaskResource


class Method(object):
    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    DELETE = 'DELETE'


class Status(object):
    OK = 200
    CREATED = 201
    NO_CONTENT = 204


class Resource(FlaskResource):
    endpoint = ''
    serializer = None

    def __init__(self):
        self._serializer = self.serializer()

    def post(self):
        return self._serializer.create(request.json), Status.CREATED

    def put(self, id):
        return self._serializer.update(id, request.json), Status.OK

    def delete(self, id):
        return self._serializer.delete(id), Status.NO_CONTENT

    def get(self, id=None):
        return self._serializer.read(id) if id else self._serializer.list(
            request.args.to_dict()
        )


def add_resource(api, resource, identifier_type='int'):
    endpoint = resource.endpoint

    api.add_resource(
        resource,
        '/{}'.format(endpoint),
        methods=[Method.GET, Method.POST],
        endpoint='{}-list'.format(endpoint)
    )

    api.add_resource(
        resource,
        '/{}/<{}:id>'.format(endpoint, identifier_type),
        methods=[Method.GET, Method.PUT, Method.DELETE],
        endpoint=endpoint + '-detail'
    )



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


setup(
    name='simplyrestful',
    version='0.3.2',
    keywords='Simply ReSTful REST API',
    license='MIT',
    platforms='all',
    author='Gabriel Jose Bazan',
    author_email='gbazan@outlook.com',
    maintainer='Gabriel Jose Bazan',
    maintainer_email='gbazan@outlook.com',
    include_package_data=True,
    url='https://github.com/gabrielbazan/simply-restful/',
    download_url='https://github.com/gabrielbazan/simply-restful/',
    description='A simple framework to quickly implement ReSTful APIs.',
    long_description=open('README.txt').read(),
    packages=[
        'simplyrestful',
        'simplyrestful/models'
    ],
    install_requires=[
        'flask',
        'flask_restful',
        'sqlalchemy',
        'geoalchemy2',
        'psycopg2',
        'shapely',
        'geojson'
    ]
)



from simplyrestful.database import Base


class Model(Base):
    __abstract__ = True

    @property
    def relationships(self):
        return {
            relationship.local_remote_pairs[0][0].name: relationship.key
            for relationship in self.__mapper__.relationships
        }

    @property
    def relationship_classes(self):
        return [
            getattr(self.__class__, t).property.mapper.class_
            for f, t in self.relationships.iteritems()
        ]

    @property
    def primary_keys(self):
        return [c.name for c in self.__mapper__.primary_key]

    @property
    def columns(self):
        return [c.name for c in self.model.__table__.columns]
